#include <sys/types.h>
#include <unistd.h>
#include <stdlib.h>
#include <stdarg.h>
#include <stdio.h>
#include <string.h>
#include <errno.h>
#include <pwd.h>
#include <grp.h>
#include <sys/socket.h>
#include <netdb.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <signal.h>
#include <sys/stat.h>
#include <syslog.h>
#include <time.h>

#ifndef INADDR_NONE
# define INADDR_NONE ((in_addr_t)-1)
#endif

#include "rbldnsd.h"
#include "dns.h"

static unsigned ttl = 2048;
static int recheck = 60;
unsigned char buf[1024];

/* a list of zonetypes. */
static struct zonetype {
  const char *name;
  z_queryfn *queryfn;
  z_loadfn *loadfn;
  z_freefn *freefn;
} zonetypes[] = {
  /* first type is the default */
#define zonetype(t) { #t, t##_query, t##_load, t##_free }
  zonetype(simple),
#undef zonetype
  { NULL, NULL, NULL, NULL }
};

static char *progname;
static int logto;
#define LOGTO_STDOUT 0x01
#define LOGTO_SYSLOG 0x02

static void error(int errnum, const char *fmt, ...) {
  char buf[128];
  int ll = sprintf(buf, "%s: ", progname), l = ll;
  va_list ap;
  va_start(ap, fmt);
  l += vsprintf(buf + l, fmt, ap);
  if (errnum)
    l += sprintf(buf + l, ": %s", strerror(errnum));
  if (logto & LOGTO_SYSLOG)
    syslog(LOG_ERR, "%s", buf + ll);
  buf[l++] = '\n';
  write(2, buf, l);
  exit(1);
}

void log(int level, const char *fmt, ...) {
  va_list ap;
  va_start(ap, fmt);
  if (logto & LOGTO_SYSLOG)
    vsyslog(level, fmt, ap);
  if (logto & LOGTO_STDOUT) {
    char buf[1024];
    int l = sprintf(buf, "%s: ", progname);
    l += vsprintf(buf + l, fmt, ap);
    buf[l++] = '\n';
    write(1, buf, l);
  }
}

struct zone {
  const char *name, *type;
  unsigned char *base;
  unsigned blen;
  z_queryfn *queryfn;
  z_loadfn *loadfn;
  z_freefn *freefn;
  time_t loaded;
  struct zonedata *data;
};

static int loadzones(struct zone *zone) {
  int r = 1;
  struct stat st;
  for(; zone->name; ++zone) {
    if (stat(zone->name, &st) < 0) {
      log(LOG_ERR, "unable to stat zonefile %s: %s",
	  zone->name, strerror(errno));
      r = 0;
      continue;
    }
    if (st.st_mtime == zone->loaded)
      continue;
    if (zone->data)
      zone->freefn(zone->data);
    if (!(zone->data = zone->loadfn(zone->name)))
      r = 0;
    else {
      zone->loaded = st.st_mtime;
      log(LOG_INFO, "zone %s loaded (%.24s)",
	  zone->name, ctime(&zone->loaded));
    }
  }
  return r;
}

static int satoi(const char *s) {
  int n = 0;
  if (*s < '0' || *s > '9') return -1;
  do n = n * 10 + (*s++ - '0');
  while (*s >= '0' && *s <= '9');
  return *s ? -1 : n;
}

static void usage() {
   printf(
"%s: rbl dns daemon\n"
"Usage is: %s [options] zone...\n"
"where options are:\n"
" -u user[:group] - run as this user:group (rbldns)\n"
" -r rootdir - chroot to this directory\n"
" -w workdir - working directory with zone files\n"
" -b [address][:port] - bind to (listen on) this address (*:53)\n"
" -t ttl - TTL value set in answers (2048)\n"
" -c check - check for file updates every `check' secs (60)\n"
" -n - do not become a daemon\n"
"each zone specified using `type:name' syntax, where type is\n"
"a type of input zone file format and name is both a name of\n"
"a zone and a name of a file to load it from.\n"
, progname, progname);
   exit(0);
}

static int init(int argc, char **argv, struct zone **zonep) {
  int c;
  char *p;
  char *user = "rbldns", *bindaddr = "";
  char *root = NULL, *workdir = NULL;
  uid_t uid = 0;
  gid_t gid = 0;
  struct sockaddr_in sin;
  int fd;
  struct zone *z;
  int nodaemon = 0;

  if ((progname = strrchr(argv[0], '/')) != NULL)
    argv[0] = ++progname;
  else
    progname = argv[0];

  if (argc <= 1) usage();

  while((c = getopt(argc, argv, "u:r:b:w:t:c:nh")) != EOF)
    switch(c) {
    case 'u': user = optarg; break;
    case 'r': root = optarg; break;
    case 'b': bindaddr = optarg; break;
    case 'w': workdir = optarg; break;
    case 't':
      if ((c = satoi(optarg)) < 0)
        error(0, "invalid ttl (-t) value `%s'", optarg);
      ttl = c;
      break;
    case 'c':
      if ((c = satoi(optarg)) < 0)
        error(0, "invalid check interval (-c) value `%s'", optarg);
      recheck = c;
      break;
    case 'n': nodaemon = 1; break;
    case 'h': usage(); break;
    default: error(0, "type `%s -h' for help", progname);
    }

  if (!(argc -= optind))
    error(0, "no zone(s) to service specified (-h for help)");
  argv += optind;

  if (nodaemon)
    logto = LOGTO_STDOUT;
  else {
    tzset();
    openlog(progname, LOG_PID|LOG_NDELAY, LOG_DAEMON);
    logto = LOGTO_STDOUT | LOGTO_SYSLOG;
  }

  fd = socket(PF_INET, SOCK_DGRAM, IPPROTO_UDP);
  if (fd < 0)
    error(errno, "unable to create listening socket");
  memset(&sin, 0, sizeof(sin));
  sin.sin_family = AF_INET;
  if (bindaddr) {
    if ((p = strchr(bindaddr, ':')) != NULL)
      *p++ = '\0';
    if (*bindaddr &&
        (sin.sin_addr.s_addr = inet_addr(bindaddr)) == INADDR_NONE) {
      struct hostent *he = gethostbyname(bindaddr);
      if (!he)
        error(0, "invalid bind address specified: `%s'", bindaddr);
      if (he->h_addrtype != AF_INET || he->h_length != sizeof(sin.sin_addr))
        error(0, "unexpected type of listening address `%s'", bindaddr);
      memcpy(&sin.sin_addr, he->h_addr_list[0], sizeof(sin.sin_addr));
      endhostent();
    }
  }
  else
    p = NULL;
  if (p && *p) {
    if ((c = satoi(p)) > 0)
      ;
    else {
      struct servent *se = getservbyname(p, "udp");
      if (!se)
        error(0, "%s/udp: unknown service", p);
      c = se->s_port;
      endservent();
    }
    p[-1] = ':';
  }
  else
    c = DNS_PORT;
  sin.sin_port = htons(c);
  if (bind(fd, (struct sockaddr *)&sin, sizeof(sin)) < 0)
    error(errno, "unable to bind to %s:%d", inet_ntoa(sin.sin_addr), c);

  c = 65536;
  do
    if (setsockopt(fd, SOL_SOCKET, SO_RCVBUF, &c, sizeof c) == 0)
      break;
  while ((c -= (c >> 5)) >= 1024);

  if ((p = strchr(user, ':')) != NULL)
    *p++ = '\0';
  if ((c = satoi(user)) >= 0)
    uid = c, gid = c;
  else {
    struct passwd *pw = getpwnam(user);
    if (!pw)
      error(0, "invalid user `%s' specified", user);
    uid = pw->pw_uid;
    gid = pw->pw_gid;
    endpwent();
  }
  if (!uid)
    error(0, "daemon should not run as root, specify -u option");
  if (p) {
    if ((c = satoi(p)) >= 0)
      gid = c;
    else {
      struct group *gr = getgrnam(p);
      if (!gr)
        error(0, "invalid group `%s' specified", p);
      gid = gr->gr_gid;
      endgrent();
    }
    p[-1] = ':';
  }

  if (root && (chdir(root) < 0 || chroot(root) < 0))
    error(errno, "unable to chroot to %s", root);
  if (workdir && chdir(workdir) < 0)
    error(errno, "unable to chdir to %s", workdir);

  if (setgroups(1, &gid) < 0 || setgid(gid) < 0 || setuid(uid) < 0)
    error(errno, "unable to setuid(%d:%d)", uid, gid);

  z = (struct zone*)malloc(sizeof(struct zone) * (argc + 1));
  if (!z)
    error(0, "out of memory");
  memset(z, 0, sizeof(struct zone) * (argc + 1));
  for(c = 0; c < argc; ++c) {
    unsigned i;
    char *name = strchr(argv[c], ':');
    if (name) {
      p = argv[c];
      *name++ = '\0';
      for(i = 0;; ++i)
	if (!zonetypes[i].name)
	  error(0, "unknown zone type `%s'", p);
	else if (strcmp(p, zonetypes[i].name) == 0)
	  break;
    }
    else {
      name = argv[c];
      i = 0;
      p = (char*)zonetypes[i].name;
    }
    z[c].type = zonetypes[i].name;
    z[c].queryfn = zonetypes[i].queryfn;
    z[c].loadfn = zonetypes[i].loadfn;
    z[c].freefn = zonetypes[i].freefn;
    z[c].name = name;
    if (p == argv[c]) name[-1] = ':';
    i = dns_ptodn(argv[c], buf, sizeof(buf));
    if (!i)
      error(errno, "invalid zone name `%s'", argv[c]);
    z[c].blen = i;
    if (!(z[c].base = (unsigned char*)malloc(i)))
      error(0, "out of memory");
    memcpy(z[c].base, buf, i);
  }
  loadzones(z);
  *zonep = z;

  if (!nodaemon) {
    if (fork() > 0) exit(0);
    close(0); close(1); close(2);
    setsid();
    logto = LOGTO_SYSLOG;
  }
  log(LOG_INFO, "started");

  return fd;
}

/* DNS packet:
 * 0:1   identifier (client supplied)
 * 2     flags1
 *  0x80 QR response flag
 *  0x78 opcode: 0 query
 *  0x04 AA auth answer
 *  0x02 TC truncation flag
 *  0x01 RD recursion desired (may be set in query)
 * 3     flags2
 *  0x80 RA recursion available
 *  0x70 Z reserved
 *  0x0f RCODE:
 *     0 ok, 1 format error, 2 servfail, 3 nxdomain, 4 notimpl, 5 refused
 * 4:5   qdcount (numqueries)
 * 6:7   ancount (numanswers)
 * 8:9   nscount (numauthority)
 * 10:11 arcount (numadditional)
 */

static unsigned process(unsigned char *const p, unsigned char *const e,
		        const struct zone *zone) {
  unsigned char *t = p + 12; /* after header */
  unsigned char *q, *x;
  unsigned c, qlen, qcls, qtyp, qflg;
  unsigned char query[DNS_MAXDN+1];
  if (t >= e) return 0; /* short packet */
  if (p[2] & 0x80) return 0; /* response?! */
  if (p[4] || p[5] != 1) return 0; /* qdcount != 1 */
  /* check query's DN */
  for(q = query;;) {
    if ((c = *q++ = *t++) == 0)
      break;
    if (c > DNS_MAXLABEL) return 0; /* no compressed names expected */
    if (t + c >= e || q + c - query > DNS_MAXDN) return 0;
    while(c--)
      *q++ = (*t >= 'A' && *t <= 'Z') ? *t++ - 'A' + 'a' : *t++;
  }
  qlen = q - query;

  if (t + 4 > e) return 0;
  qtyp = ((unsigned)(t[0]) << 8) | t[1];
  qcls = ((unsigned)(t[2]) << 8) | t[3];
  t += 4;

  /* from now on, we see (almost?) valid dns query, should reply */

  /* construct reply packet */
  /* identifier already in place */
  /* flags will be set up later */
  /* p[4:5] (qdcount) already set up in query */
  p[6] = p[7] = 0; /* ancount */
  p[8] = p[9] = 0; /* nscount */
  p[10] = p[11] = 0; /* arcount */

  if (qcls != DNS_C_IN && qcls != DNS_C_ANY)
    return refuse(p,t,1); /* format error */
  if (p[2] & 126)
    return refuse(p,t,4); /* not implemented */
  if (qtyp == DNS_T_ANY) qflg = QUERY_A|QUERY_TXT;
  else if (qtyp == DNS_T_A) qflg = QUERY_A;
  else if (qtyp == DNS_T_TXT) qflg = QUERY_TXT;
  else
    return refuse(p,t,5); /* refused */

  p[2] = 0x80; /* 0x81?! */
  if (qcls == DNS_C_IN) p[2] |= 0x04; /* AA */
  p[3] = 0;
  for(; zone->name; ++zone) {
    if (!zone->data) continue;
    if (zone->blen > qlen) continue;
    x = query + qlen - zone->blen;
    if (memcmp(zone->base, x, zone->blen)) continue;
    q = query;
    while(q < x) q += *q + 1;
    if (q != x) continue;
    *q = 0;
    c = zone->queryfn(zone->data, p, t, query, qflg);
    *q = zone->base[0];
    if (c) return c;
  }

  return refuse(p,t,5); /* refused */
}

static volatile int reload;
static void sighandler(int sig) {
  reload = 1;
  switch(sig) {
  case SIGALRM:
    alarm(recheck);
    reload = 1;
    break;
  case SIGHUP:
    log(LOG_INFO, "reloading");
    reload = 1;
    break;
  case SIGTERM:
  case SIGINT:
    log(LOG_INFO, "terminating");
    exit(0);
  }
}

static void setup_signals() {
#if 1
   struct sigaction sa;
   memset(&sa, 0, sizeof(sa));
   sa.sa_handler = sighandler;
# define setsig(sig) sigaction(sig, &sa, NULL)
#else
# define setsig(sig) signal(sig, sighandler)
#endif
   setsig(SIGHUP);
   setsig(SIGALRM);
   setsig(SIGTERM);
   setsig(SIGINT);
#undef setsig
}

int main(int argc, char **argv) {
  int fd;
  struct zone *zone;
  fd = init(argc, argv, &zone);
  setup_signals();
  alarm(recheck);
  for(;;) {
    struct sockaddr_in sin;
    socklen_t sinl = sizeof(sin);
    int r;
    if (reload) {
      loadzones(zone);
      reload = 0;
    }
    r = recvfrom(fd, buf, DNS_MAXPACKET, 0, (struct sockaddr *)&sin, &sinl);
    if (r < 0)
      continue;
    if (!(r = process(buf, buf + r, zone)))
      continue;
    while(sendto(fd, buf, r, 0, (struct sockaddr *)&sin, sizeof(sin)) < 0 &&
          errno == EINTR);
  }
  return 0;
}

unsigned char *
addrec_a(unsigned char *const p, unsigned char *t, in_addr_t aip) {
  if (t - p + 16 > DNS_MAXPACKET)
    return t;
  p[7]++; /* numanswers */
  *t++ = 192; *t++ = 12;  /* jump after header */
  *t++ = DNS_T_A>>8; *t++ = DNS_T_A; /* type */
  *t++ = DNS_C_IN>>8; *t++ = DNS_C_IN; /* class */
  *t++ = ttl>>24; *t++ = ttl>>16; *t++ = ttl>>8; *t++ = ttl;
  *t++ = 0; *t++ = 4;
  memcpy(t, &aip, 4);
  return t + 4;
}

unsigned char *
addrec_txt(unsigned char *const p, unsigned char *t, const char *txt,
	   in_addr_t qip) {
  unsigned char *lp;
  unsigned char *e = p + DNS_MAXPACKET;
  char ipt[16];
  unsigned iptl = 0;
  if (t + 11 > e)
    return t;
  else if (e - t > 255 + 11) e = t + 255 + 11;
  p[7]++; /* ancount */
  *t++ = 192; *t++ = 12; /* jump to after header */
  *t++ = DNS_T_TXT>>8; *t++ = DNS_T_TXT;
  *t++ = DNS_C_IN>>8; *t++ = DNS_C_IN;
  *t++ = ttl>>24; *t++ = ttl>>16; *t++ = ttl>>8; *t++ = ttl;
  *t++ = 0;
  lp = t; /* length pointer */
  t += 2;
  while(*txt && t < e) {
    if (*txt == '$') {
      if (!iptl)
	iptl =
	  sprintf(ipt, "%d.%d.%d.%d",
		  qip&255, (qip>>8)&255, (qip>>16)&255, (qip>>24));
      if (t + iptl > e) iptl = e - t;
      memcpy(t, ipt, iptl);
      t += iptl;
      ++txt;
    }
    else
      *t++ = *txt++;
  }
  lp[0] = t - lp - 1;
  lp[1] = t - lp - 2;
  return t;
}

int dn4octet(const unsigned char *q, in_addr_t *addrp) {
  unsigned x, j;
  in_addr_t a = 0;
  for(j = 0;; ++j) {
    if (!*q) { *addrp = a; return 1; }
    if (j >= 4) return 0;
    if (*q >= 4) return 0;
    if (q[1] < '0' || q[1] > '9') return 0;
    x = q[1] - '0';
    if (*q == 1) { a = a << 8 | x; q += 2; continue; }
    if (!x) return 0;
    if (q[2] < '0' || q[2] > '9') return 0;
    x = x * 10 + q[2] - '0';
    if (*q == 2) { a = a << 8 | x; q += 3; continue; }
    if (q[3] < '0' || q[3] > '9') return 0;
    x = x * 10 + q[3] - '0';
    if (x > 255) return 0;
    a = a << 8 | x;
    q += 4;
  }
}

void nomem(const char *zone, unsigned size) {
  log(LOG_ERR, "out of memory while loading zone %s (%d bytes)", zone, size);
}

void *emalloc(const char *zone, unsigned size) {
  void *ptr = malloc(size);
  if (!ptr)
    nomem(zone, size);
  return ptr;
}

void *erealloc(const char *zone, void *ptr, unsigned size) {
  void *nptr = realloc(ptr, size);
  if (!nptr) {
    free(ptr);
    nomem(zone, size);
  }
  return nptr;
}

char *estrdup(const char *zone, const char *str) {
  char *s = strdup(str);
  if (!s)
    nomem(zone, strlen(str) + 1);
  return s;
}
