/* a simples zone type: a list of IP addresses with
 * a common text (maybe with a substitution for IP).
 * A list is stored as a sorted array, with binary search --
 * maybe not the most efficient way, but it's the most optimal
 * from the memory requiriments point of view, simple to
 * handle and sufficient enouth (imho)
 */

#include <netinet/in.h>
#include <arpa/inet.h>
#include <syslog.h>
#include <errno.h>
#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include "rbldnsd.h"
#include "dns.h"

#ifndef INADDR_NONE
# define INADDR_NONE ((in_addr_t)-1)
#endif

struct zonedata {
  in_addr_t a;
  char *txt;
  unsigned naddr;
  in_addr_t *addr;
};

unsigned
simple_query(const struct zonedata *const z,
	     unsigned char *const p, unsigned char *t,
	     const unsigned char *const query, unsigned qflg)
{
   int a = 0, b = z->naddr - 1, m;
   in_addr_t addr;
   if (!dn4octet(query, &addr))
     return nxdomain(p,t);
   while(a <= b) {
     m = (a + b) >> 1;
     if (z->addr[m] == addr) {
       if (qflg & QUERY_A) t = addrec_a(p, t, z->a);
       if (qflg & QUERY_TXT) t = addrec_txt(p, t, z->txt, addr);
       return t - p;
     }
     else if (z->addr[m] < addr) a = m + 1;
     else b = m - 1;
   }
   return nxdomain(p, t);
}

static int cmpaddr(const void *a, const void *b) {
  if (*(in_addr_t*)a < *(in_addr_t*)b) return -1;
  if (*(in_addr_t*)a > *(in_addr_t*)b) return 1;
  return 0;
}

struct zonedata *simple_load(const char *name) {
  struct zonedata *z;
  int lineno = 0;
  int l;
  in_addr_t *addr = NULL;
  unsigned naddr = 0, nalc = 0;
  unsigned a;
  FILE *f;
  if ((f = fopen(name, "r")) == NULL) {
    log(LOG_ERR, "unable to open zone file `%s': %s", name, strerror(errno));
    return NULL;
  }
  z = (struct zonedata *)emalloc(name, sizeof(*z));
  if (!z) {
    fclose(f);
    return NULL;
  }
  z->a = inet_addr("127.0.0.2");
  z->txt = estrdup(name, "listed");
  if (!z->txt) {
    free(z);
    fclose(f);
    return NULL;
  }

  while(fgets(buf, sizeof(buf), f)) {
    ++lineno;
    l = strlen(buf);
    if (l && buf[l-1] == '\n') buf[--l] = '\0';
    if (!l || buf[0] == '#') continue;
    if (lineno == 1 && buf[0] == ':') {
      char *p = strchr(buf+1, ':');
      if (p) {
        *p++ = '\0';
        if ((a = inet_addr(buf+1)) != INADDR_NONE)
          z->a = a;
      }
      else
        p = buf + 1;
      if (*p) {
	free(z->txt);
        if ((z->txt = estrdup(name, p)) == NULL) {
	  free(z);
	  fclose(f);
	  return NULL;
	}
      }
      continue;
    }
    if ((a = inet_addr(buf)) == INADDR_NONE) {
      log(LOG_WARNING, "zone %s, line %d: invalid entry (ignored)",
	  name, lineno);
      continue;
    }
    if (naddr == nalc) {
      nalc = nalc ? nalc << 1 : 512;
      addr = (in_addr_t *)erealloc(name,  addr, nalc * sizeof(in_addr_t));
      if (!addr) {
	free(z->txt);
	free(z);
	fclose(f);
	return NULL;
      }
    }
    addr[naddr++] = a;
  }
  fclose(f);
  addr = (in_addr_t *)realloc(addr, naddr * sizeof(in_addr_t));
  qsort(addr, naddr, sizeof(*addr), cmpaddr);
  z->addr = addr;
  z->naddr = naddr;
  return z;
}

void simple_free(struct zonedata *z) {
  if (z->txt) free(z->txt);
  if (z->addr) free(z->addr);
  free(z);
}
